"""
Base classes.
"""
import os
from functools import partial
from pathlib import Path

from maven import utils

CACHING_ENABLED = True
VERBOSE = False


class Pipeline:
    """Generic class for retrieving & processing datasets with built-in caching & MD5 checking."""

    def __init__(self, directory):
        self.directory = Path(directory)
        self.sources = []  # tuples of (url, filename, checksum)
        self.retrieve_all = False
        self.target = (None, None)
        self.verbose_name = ""
        self.year = None

    def retrieve(self):
        """Retrieve data from self.sources into self.directory / 'raw' and validate against checksum."""
        target_dir = self.directory / "raw"
        os.makedirs(target_dir, exist_ok=True)  # create directory if it doesn't exist
        for url, filename, md5_checksum in self.sources:
            if utils.is_url(url):
                processing_fn = partial(utils.fetch_url, url=url, filename=filename, target_dir=target_dir)
            else:
                processing_fn = partial(utils.get_and_copy, identifier=url, filename=filename, target_dir=target_dir)
            utils.retrieve_from_cache_if_exists(
                filename=filename,
                target_dir=target_dir,
                processing_fn=processing_fn,
                md5_checksum=md5_checksum,
                caching_enabled=CACHING_ENABLED,
                verbose=VERBOSE,
            )
            if not self.retrieve_all:  # retrieve just the first dataset
                return
        if self.retrieve_all:  # all datasets retrieved
            return
        else:  # retrieving first dataset only but all fallbacks failed
            raise RuntimeError(f"Unable to download {self.verbose_name} data.")

    def process(self):
        pass


class UKResults(Pipeline):
    """Handles results data for UK General Elections."""

    def process(self):
        """Process results data for a UK General Election."""
        filename = self.sources[0][1]
        processed_results_location = self.directory / "processed" / self.target[0]
        os.makedirs(self.directory / "processed", exist_ok=True)  # create directory if it doesn't exist

        def process_and_export():
            # Either caching disabled or file not yet processed; process regardless.
            results = utils.process_hoc_sheet(input_file=filename, data_dir=self.directory, sheet_name=str(self.year))
            # Export
            print(f"Exporting dataset to {processed_results_location.resolve()}")
            results.to_csv(processed_results_location, index=False)

        utils.retrieve_from_cache_if_exists(
            filename=self.target[0],
            target_dir=(self.directory / "processed"),
            processing_fn=process_and_export,
            md5_checksum=self.target[1],
            caching_enabled=CACHING_ENABLED,
            verbose=VERBOSE,
        )



from .uk_2010_results import UK2010Results
from .uk_2015_model import UK2015Model
from .uk_2015_results import UK2015Results
from .uk_2017_results import UK2017Results
from .uk_polls import UKPolls

__all__ = ["UK2010Results", "UK2015Model", "UK2015Results", "UK2017Results", "UKPolls"]



"""
Various helper functions.
"""

import hashlib
import shutil
import warnings
from urllib.parse import urlparse

import pandas as pd
import requests

import maven


def sanitise(x):
    return x.lower().replace(" ", "_")


def process_hoc_sheet(input_file, data_dir, sheet_name):
    # Import general election results
    print(f"Read and clean {input_file}")
    parties = ["Con", "LD", "Lab", "UKIP", "Grn", "SNP", "PC", "DUP", "SF", "SDLP", "UUP", "APNI", "Other"]
    results = pd.read_excel(
        data_dir / "raw" / input_file, sheet_name=sheet_name, skiprows=4, header=None, skipfooter=19
    )
    assert results.shape == (650, 49)

    # Specify columns (spread across multiple rows in Excel)
    cols = ["", "id", "Constituency", "County", "Country/Region", "Country", "Electorate", ""]
    for party in parties:
        cols += [f"{party}_Votes", f"{party}_Voteshare", ""]
    cols += ["Total votes", "Turnout"]
    results.columns = cols

    # Some basic data quality checks
    for party in parties:
        assert (results[f"{party}_Voteshare"] - results[f"{party}_Votes"] / results["Total votes"]).sum() == 0
    assert (results[[f"{party}_Votes" for party in parties]].fillna(0.0).sum(axis=1) == results["Total votes"]).all()
    assert ((results["Total votes"] / results["Electorate"]) == results["Turnout"]).all()

    # Drop blank columns plus those that can be calculated
    cols_to_drop = [""] + [c for c in cols if "Voteshare" in c] + ["Total votes", "Turnout"]
    results = results.drop(columns=cols_to_drop)

    # Sanitise column names
    results.columns = [sanitise(c) for c in results.columns]
    results = results.rename(columns={"id": "ons_id", "country/region": "region"})
    results.columns = [c.replace("_votes", "") for c in results.columns]

    # Reshape to long
    results_long = pd.melt(
        results,
        id_vars=["ons_id", "constituency", "county", "region", "country", "electorate"],
        var_name="party",
        value_name="votes",
    )
    assert results.shape == (650, 19)
    assert results_long.shape == (650 * len(parties), 19 - len(parties) + 2)

    # Sort by (ons_id, party)
    results_long["party"] = pd.Categorical(
        results_long.party, categories=pd.Series(parties).apply(sanitise), ordered=True
    )
    results_long = results_long.sort_values(["ons_id", "party"]).reset_index(drop=True)

    # Re-add total_votes & voteshare
    results_long["total_votes"] = results_long.ons_id.map(results_long.groupby("ons_id").votes.sum().astype(int))
    results_long["voteshare"] = results_long["votes"] / results_long["total_votes"]
    results_long["turnout"] = results_long["total_votes"] / results_long["electorate"]

    # Reorder cols for export
    results_long = results_long[
        [
            "ons_id",
            "constituency",
            "county",
            "region",
            "country",
            "electorate",
            "total_votes",
            "turnout",
            "party",
            "votes",
            "voteshare",
        ]
    ].copy()

    return results_long


def calculate_md5_checksum(filename):
    """
    Calculate the checksum of the file, exactly same as md5-sum linux util.
    Code from https://github.com/RaRe-Technologies/gensim/blob/develop/gensim/downloader.py
    """
    hash_md5 = hashlib.md5()
    with open(filename, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            hash_md5.update(chunk)
    return hash_md5.hexdigest()


def is_url(url):
    """Source: https://stackoverflow.com/a/52455972"""
    try:
        result = urlparse(url)
        return all([result.scheme, result.netloc])
    except ValueError:
        return False


def fetch_url(url, filename, target_dir):
    """Download filename from url into target_dir."""
    response = requests.get(url + filename)
    if response.status_code != 200:
        warnings.warn(f"Received status {response.status_code} when trying to retrieve {url}{filename}")
    # Save to file
    with open(target_dir / filename, "wb") as f:
        f.write(response.content)
    print(f"Successfully downloaded {filename} into {target_dir.resolve()}")
    return target_dir / filename


def get_and_copy(identifier, filename, target_dir):
    # """Run maven.get(identifier) and copy filename from identifier/processed/ data into target/."""
    # target_dir by default is data/general-election/UK/2015/model
    subdirectories_below = str(target_dir).count("/")
    go_up = "/".join([".." for _ in range(subdirectories_below)])
    data_directory = (target_dir / go_up).resolve()  # sensible guess?
    maven.get(identifier, data_directory=data_directory)
    source = data_directory / identifier / "processed"
    print(f"Copying {filename} from {source} -> {target_dir}.")
    shutil.copyfile(src=source / filename, dst=target_dir / filename)


def retrieve_from_cache_if_exists(
    filename, target_dir, processing_fn, md5_checksum=None, caching_enabled=True, verbose=False
):
    """Retrieve filename from target_dir if it exists, otherwise execute processing_fn.

    Raises a warning if the retrieved/processed file's checksum doesn't match the expected MD5.
    """
    if caching_enabled and (target_dir / filename).exists():
        # Check if it's already in target_dir.
        print(f"Cached file {filename} is already in {target_dir.resolve()}")
    else:
        # Either caching disabled or file not there yet.
        processing_fn()

    # File should now be there. Let's check checksums.
    downloaded_file_md5_checksum = calculate_md5_checksum(target_dir / filename)
    if verbose:
        print(f"Checksum for {filename}: {downloaded_file_md5_checksum}")
    if md5_checksum and downloaded_file_md5_checksum != md5_checksum:
        warnings.warn(f"MD5 checksum doesn't match for {filename}")


##########################
# ELECTION MODELLING UTILS
##########################


def calculate_winners(df, voteshare_col):
    """Assumes df has `ons_id` and `party` columns."""
    return (
        df.sort_values(voteshare_col, ascending=False)
        .groupby("ons_id")
        .head(1)[["ons_id", "party"]]
        .set_index("ons_id")
        .party
    )



"""
Results data for the United Kingdom's 2017 General Election.

Usage:
    >>> import maven
    >>> maven.get('general-election/UK/2017/results', data_directory='./data/')


Sources:
    - http://researchbriefings.files.parliament.uk/documents/CBP-8647/1918-2017election_results.csv
        - From https://researchbriefings.parliament.uk/ResearchBriefing/Summary/CBP-8647

Other sources:
    - https://researchbriefings.parliament.uk/ResearchBriefing/Summary/CBP-7186
    - http://researchbriefings.files.parliament.uk/documents/CBP-7979/HoC-GE2017-constituency-results.csv
"""

from pathlib import Path

from maven.datasets.general_election.base import UKResults


class UK2017Results(UKResults):
    """Handles results data for the United Kingdom's 2017 General Election."""

    def __init__(self, directory=Path("data/general-election/UK/2017/results")):
        super(UK2017Results, self).__init__(directory=directory)
        self.directory = Path(directory)
        self.sources = [
            # url, filename, checksum
            (
                "http://researchbriefings.files.parliament.uk/documents/CBP-8647/",
                "1918-2017election_results_by_pcon.xlsx",
                "a1e4628945574639b541b21bada2531c",
            ),
        ]
        self.target = ("general_election-uk-2017-results.csv", "c7e1fde647e55f9d4567cb81e62c782a")  # filename, checksum
        self.verbose_name = "UK 2017 General Election results"
        self.year = "2017"



"""
Main data getting functionality. Maps data identifiers to data pipeline classes.

Example usage:
    > import maven
    > maven.get('general-election/UK/2015/results', data_directory='./data/')
"""

from pathlib import Path

from .datasets import general_election


def get(name, data_directory=Path("."), retrieve=True, process=True):
    """Core data getter function.

    Args:
        name (str): Name of dataset to retrieve/process.
        data_directory (str or pathlib.PosixPath): Path to directory where datasets will be saved (either as string
                                                   a pathlib Path).
        retrieve (bool): Toggle dataset retrieval.
        process (bool): Toggle dataset processing.

    Returns: Nothing (datasets are placed into current working directory).
    """
    mapper = {
        "general-election/UK/2010/results": general_election.UK2010Results,
        "general-election/UK/2015/model": general_election.UK2015Model,
        "general-election/UK/2015/results": general_election.UK2015Results,
        "general-election/UK/2017/results": general_election.UK2017Results,
        "general-election/UK/polls": general_election.UKPolls,
    }
    if name not in mapper:
        raise KeyError(f"'{name}' not found in datasets.")

    if isinstance(data_directory, str):
        data_directory = Path(data_directory)
    pipeline = mapper[name](directory=(data_directory / name))

    if retrieve:
        pipeline.retrieve()
    if process:
        pipeline.process()



"""
Results data for the United Kingdom's 2015 General Election.

Usage:
    >>> import maven
    >>> maven.get('general-election/UK/2015/results', data_directory='./data/')


Sources:
    - http://researchbriefings.files.parliament.uk/documents/CBP-8647/1918-2017election_results.csv
        - From https://researchbriefings.parliament.uk/ResearchBriefing/Summary/CBP-8647

Deprecated sources:
    - http://www.electoralcommission.org.uk/__data/assets/file/0004/191650/2015-UK-general-election-data-results-WEB.zip

Notes:
    - 2015-UK-general-election-data-results-WEB.zip has a lot more detailed data.
"""

from pathlib import Path

from maven.datasets.general_election.base import UKResults


class UK2015Results(UKResults):
    """Handles results data for the United Kingdom's 2015 General Election."""

    def __init__(self, directory=Path("data/general-election/UK/2015/results")):
        super(UK2015Results, self).__init__(directory=directory)
        self.directory = Path(directory)
        self.sources = [
            # url, filename, checksum
            (
                "http://researchbriefings.files.parliament.uk/documents/CBP-8647/",
                "1918-2017election_results_by_pcon.xlsx",
                "a1e4628945574639b541b21bada2531c",
            ),
        ]
        self.target = ("general_election-uk-2015-results.csv", "9a785cb19275e4dbc79da67eece6067f")  # filename, checksum
        self.verbose_name = "UK 2015 General Election results"
        self.year = "2015"



"""
Model-ready dataset for the United Kingdom's 2015 General Election.

Usage:
    > import maven
    > maven.get('general-election/UK/2015/model', data_directory='./data/')
"""
import os
from pathlib import Path

import pandas as pd

from maven import utils
from maven.datasets.general_election.base import Pipeline


class UK2015Model(Pipeline):
    """Generates model-ready data for the United Kingdom's 2015 General Election."""

    def __init__(self, directory=Path("data/general-election/UK/2015/model")):
        super(UK2015Model, self).__init__(directory=directory)  # inherit base __init__ but override default directory
        self.sources = [
            # tuples of (url, filename, checksum)
            (
                "general-election/UK/2010/results",
                "general_election-uk-2010-results.csv",
                "954a0916f5ce791ca566484ce566088d",
            ),
            (
                "general-election/UK/2015/results",
                "general_election-uk-2015-results.csv",
                "9a785cb19275e4dbc79da67eece6067f",
            ),
            ("general-election/UK/polls", "general_election-uk-polls.csv", "98f865803c782e1ffd0cdc4774707ae5"),
            ("general-election/UK/polls", "general_election-london-polls.csv", "97eb4254039a6bca1a882a9afde2b067"),
            ("general-election/UK/polls", "general_election-scotland-polls.csv", "096354c852a7c30e22a733eec133b9e3"),
            ("general-election/UK/polls", "general_election-wales-polls.csv", "2134d55e5288bd5b12be2471f4aacab7"),
            ("general-election/UK/polls", "general_election-ni-polls.csv", "ea871fad0ce51c03dda09ecec0377dc6"),
        ]
        self.retrieve_all = True
        self.verbose_name = "UK2015Model"
        self.year = 2015

    def process(self):
        """Process results data from the United Kingdom's 2010 and 2015 General Elections
            into a single model-ready dataset for predicting the 2015 General Election."""
        processed_directory = self.directory / "processed"
        os.makedirs(processed_directory, exist_ok=True)  # create directory if it doesn't exist

        #############
        # IMPORT DATA
        #############
        train = 2010
        test = 2015
        pred = 2017
        # geos sit between region and country (e.g. "england_not_london") and map to things we can extract from polls
        geos = ["uk", "scotland", "wales", "ni", "london"]
        geo_lookup = {
            "Northern Ireland": "ni",
            "Scotland": "scotland",
            "Wales": "wales",
            "London": "london",
            "South East": "england_not_london",
            "West Midlands": "england_not_london",
            "North West": "england_not_london",
            "East Midlands": "england_not_london",
            "Yorkshire and The Humber": "england_not_london",
            "Eastern": "england_not_london",
            "South West": "england_not_london",
            "North East": "england_not_london",
        }
        results_seat_count = {
            2010: {
                "con": 306,
                "lab": 258,
                "ld": 57,
                "dup": 8,
                "snp": 6,
                "sf": 5,
                "pc": 3,
                "sdlp": 3,
                "grn": 1,
                "apni": 1,
                "other": 2,  # {'speaker': 'John Bercow', 'independent': 'Sylvia Hermon'}
            },
            2015: {
                "con": 330,
                "lab": 232,
                "snp": 56,
                "ld": 8,
                "dup": 8,
                "sf": 4,
                "pc": 3,
                "sdlp": 3,
                "uup": 2,
                "ukip": 1,
                "grn": 1,
                "other": 2,  # {'speaker': 'John Bercow', 'independent': 'Sylvia Hermon'}
            },
        }
        winner_fixes = {
            2010: [
                # https://en.wikipedia.org/wiki/Fermanagh_and_South_Tyrone_(UK_Parliament_constituency)
                ("N06000007", "sf"),  # SF = 21,304, Independent Unionist (with DUP support) = 21,300, Independent = 188
            ]
        }

        # Import general election results
        results = {}
        results[train] = pd.read_csv(self.directory / "raw" / f"general_election-uk-{train}-results.csv")
        results[test] = pd.read_csv(self.directory / "raw" / f"general_election-uk-{test}-results.csv")
        # Add geos
        results[train]["geo"] = results[train].region.map(geo_lookup)
        results[test]["geo"] = results[test].region.map(geo_lookup)
        # ge_2010 = pd.read_csv(self.directory / "raw" / "general_election-uk-2010-results.csv")
        # ge_2015 = pd.read_csv(self.directory / "raw" / "general_election-uk-2015-results.csv")
        polls = {}
        for geo in geos:
            polls[geo] = pd.read_csv(
                self.directory / "raw" / f"general_election-{geo}-polls.csv", parse_dates=["to"]
            ).sort_values("to")

        # Check constituencies are mergeable
        assert (results[train].sort_values("ons_id").ons_id == results[test].sort_values("ons_id").ons_id).all()

        # Add the winner for the results
        for year in [train, test]:
            res = results[year].copy()
            winners = utils.calculate_winners(res, "voteshare")
            res["winner"] = res.ons_id.map(winners)

            # Apply fixes
            if year in winner_fixes:
                for ons_id, actual_winner in winner_fixes[year]:
                    res.loc[res.ons_id == ons_id, "winner"] = actual_winner

            # Check this matches the results on record
            seat_count = res[["ons_id", "winner"]].drop_duplicates().groupby("winner").size()
            assert dict(seat_count) == results_seat_count[year]

            # Add boolean per row for if this party won this seat
            res["won_here"] = res.party == res.winner
            results[year] = res.copy()

        #########
        # POLLING
        #########

        election_dates = {d.year: d for d in [pd.to_datetime(x) for x in ["2010-05-06", "2015-05-07", "2017-06-08"]]}
        for year in [test, pred]:
            election_day = election_dates[year]
            one_week_before = election_day - pd.Timedelta(days=7)
            one_month_before = election_day - pd.Timedelta(days=30)

            # TODO: Remove if not needed
            # # Look at pollsters active in final month before each election
            # pollsters = polls["uk"][
            #     (polls["uk"].to >= one_month_before) & (polls["uk"].to < election_day)
            # ].company.unique()

            # Use single last poll from each pollster in final week of polling then average out
            final_polls = {}
            for geo in geos:
                final_polls[geo] = (
                    polls[geo][(polls[geo].to >= one_week_before) & (polls[geo].to < election_day)]
                    .groupby("company")
                    .tail(1)
                )

            # Calculate regional polling
            if year == 2015:
                parties = ["con", "lab", "ld", "ukip", "grn"]
                # Create new polls dictionary by geo containing simple average across all pollsters
                national_polling = final_polls["uk"].mean().loc[parties]
                # We don't yet have regional polling in 2015 for Scotland, Wales, NI, London - add as other.
                national_polling["other"] = 1 - national_polling.sum()
                poll_of_polls = {"uk": national_polling}

                # Export
                polls_csv_list = []
                for geo in poll_of_polls:
                    polls_csv_list.append(
                        pd.DataFrame(
                            {"geo": geo, "party": poll_of_polls[geo].index, "voteshare": poll_of_polls[geo]}
                        ).reset_index(drop=True)
                    )
                polls_csv = pd.concat(polls_csv_list, axis=0)
                # polls_csv.to_csv(processed_directory / f"final_polls_{year}.csv", index=False)
            elif year == 2017:
                parties = {
                    "uk": ["con", "lab", "ld", "ukip", "grn", "snp"],
                    "scotland": ["con", "lab", "ld", "ukip", "grn", "snp"],
                    "wales": ["con", "lab", "ld", "ukip", "grn", "pc"],
                    "ni": ["con", "ukip", "grn"],
                    "london": ["con", "lab", "ld", "ukip", "grn"],
                    "england_not_london": ["con", "lab", "ld", "ukip", "grn"],
                }
                all_parties = set([x for y in parties.values() for x in y])
                poll_of_polls = {}
                for geo in geos:
                    sample_size_weights = final_polls[geo].sample_size / final_polls[geo].sample_size.sum()
                    weighted_poll_of_polls = (
                        final_polls[geo][parties[geo]]
                        .multiply(sample_size_weights, axis=0)
                        .sum()
                        .reindex(all_parties, fill_value=0.0)
                    )
                    poll_of_polls[geo] = weighted_poll_of_polls

                # Estimate polling for England excluding London
                # survation_wts from http://survation.com/wp-content/uploads/2017/06/Final-MoS-Post-BBC-Event-Poll-020617SWCH-1c0d4h9.pdf
                survation_wts = pd.Series({"scotland": 85, "england": 881, "wales": 67, "ni": 16})
                survation_wts["uk"] = survation_wts.sum()
                survation_wts["london"] = 137
                survation_wts["england_not_london"] = survation_wts.england - survation_wts.london

                england_not_london = poll_of_polls["uk"] * survation_wts["uk"]
                for geo in ["scotland", "wales", "ni", "london"]:
                    england_not_london = england_not_london.sub(poll_of_polls[geo] * survation_wts[geo], fill_value=0.0)
                england_not_london /= survation_wts["england_not_london"]
                england_not_london.loc[["pc", "snp"]] = 0.0
                poll_of_polls["england_not_london"] = england_not_london

                # Fix PC (Plaid Cymru) for UK
                poll_of_polls["uk"]["pc"] = poll_of_polls["wales"]["pc"] * survation_wts["wales"] / survation_wts["uk"]

                # Add Other
                for geo in geos + ["england_not_london"]:
                    poll_of_polls[geo]["other"] = 1 - poll_of_polls[geo].sum()

                # Export
                polls_csv_list = []
                for geo in poll_of_polls:
                    polls_csv_list.append(
                        pd.DataFrame(
                            {"geo": geo, "party": poll_of_polls[geo].index, "voteshare": poll_of_polls[geo]}
                        ).reset_index(drop=True)
                    )
                polls_csv = pd.concat(polls_csv_list, axis=0)
                # polls_csv.to_csv(processed_directory / f"final_polls_{year}.csv", index=False)

            # If year is 2015, last_election is 2010
            last_election = train if year == test else test
            # Use res for code readability and write back into results[last_election] at the end
            res = results[last_election].copy()

            # Merge into previous election's results to calculate swing
            res = (
                res.merge(
                    right=polls_csv.query('geo == "uk"')[["party", "voteshare"]].rename(
                        columns={"voteshare": "national_polls"}
                    ),
                    on="party",
                    how="outer",
                )
                .sort_values(["ons_id", "party"])
                .reset_index(drop=True)
            )
            # If we have geo-polls, add those too
            if list(poll_of_polls.keys()) != ["uk"]:
                res = (
                    res.merge(
                        right=polls_csv.query('geo != "uk"')[["geo", "party", "voteshare"]].rename(
                            columns={"voteshare": "geo_polls"}
                        ),
                        on=["geo", "party"],
                        how="outer",
                    )
                    .sort_values(["ons_id", "party"])
                    .reset_index(drop=True)
                )

            #############################
            # Calculate uplifts ("swing")
            #############################

            # Calculate national voteshare
            national_voteshare_by_party = res.groupby("party").votes.sum() / res.votes.sum()
            res["national_voteshare"] = res.party.map(national_voteshare_by_party)

            # Calculate swing between last election results and latest poll-of-polls
            res["national_swing"] = (res.national_polls / res.national_voteshare) - 1

            # Forecast is previous result multiplied by swing uplift
            res["national_swing_forecast"] = res.voteshare * (1 + res.national_swing)

            # Predict the winner in each constituency using national_swing_forecast
            # Note: these are pointless for NI as polls/swings are all aggregated under "other" but results
            # are given per major party.
            national_swing_winners = utils.calculate_winners(res, "national_swing_forecast")
            res["national_swing_winner"] = res.ons_id.map(national_swing_winners)

            ########################################################
            # Calculate Geo-Level Voteshare + Swing inc. all parties
            ########################################################

            if "geo_polls" in res.columns:
                # Calculate geo-level voteshare
                votes_by_geo = res.groupby("geo").votes.sum().reset_index()
                votes_by_geo_by_party = (
                    res.groupby(["geo", "party"])
                    .votes.sum()
                    .reset_index()
                    .merge(votes_by_geo, on="geo", how="left", suffixes=("", "_geo"))
                )
                votes_by_geo_by_party["geo_voteshare"] = votes_by_geo_by_party.votes / votes_by_geo_by_party.votes_geo
                res = res.merge(
                    votes_by_geo_by_party[["geo", "party", "geo_voteshare"]], on=["geo", "party"], how="left"
                )

                # Calculate geo-swing between last election results and latest geo-polls
                res["geo_swing"] = (res.geo_polls / res.geo_voteshare) - 1

                # Forecast is previous result multiplied by swing uplift
                res["geo_swing_forecast"] = res.voteshare * (1 + res.geo_swing)

                # Predict the winner in each constituency using geo_swing_forecast
                geo_swing_winners = utils.calculate_winners(res, "geo_swing_forecast")
                res["geo_swing_winner"] = res.ons_id.map(geo_swing_winners)

            # Put res back into results[last_election]
            results[last_election] = res.copy()

        ######################################
        # Create ML-ready dataframe and export
        ######################################

        # Use "last" and "now" in order to re-use this code for either train->test or test->pred.
        last = train
        now = test
        df_cols_now = [
            "ons_id",
            "constituency",
            "county",
            "region",
            "geo",
            "country",
            "electorate",
            "total_votes",
            "turnout",
            "party",
            "votes",
            "voteshare",
            "winner",
        ]
        df_cols_last = [
            "ons_id",
            "party",
            "total_votes",
            "turnout",
            "votes",
            "voteshare",
            "national_polls",
            "national_voteshare",
            "national_swing",
            "national_swing_forecast",
            "national_swing_winner",
            "winner",
            "won_here",
        ]
        df_cols_final_geo = []
        if "geo_polls" in results[last].columns:
            df_cols_last += ["geo_polls", "geo_voteshare", "geo_swing", "geo_swing_forecast", "geo_swing_winner"]
            df_cols_final_geo += [
                "geo_polls_now",
                "geo_voteshare_last",
                "geo_swing",
                "geo_swing_forecast",
                "geo_swing_winner",
            ]
        df_cols_final = (
            [
                # Constant per constituency
                "ons_id",
                "constituency",
                "county",
                "region",
                "geo",
                "country",
                "electorate",
                "total_votes_last",
                "turnout_last",
                # Constant per party (per constituency)
                "party",
                "votes_last",
                "voteshare_last",
                "winner_last",
                "won_here_last",
                "national_voteshare_last",
                "national_polls_now",
                "national_swing",
                "national_swing_forecast",
                "national_swing_winner",
            ]
            + df_cols_final_geo
            + [
                # Target
                "total_votes_now",
                "turnout_now",
                "votes_now",
                "voteshare_now",
                "winner_now",
            ]
        )
        df = (
            results[now][df_cols_now]
            .rename(
                columns={
                    "total_votes": "total_votes_now",
                    "turnout": "turnout_now",
                    "votes": "votes_now",
                    "voteshare": "voteshare_now",
                    "winner": "winner_now",
                }
            )
            .merge(
                # Note: even though polling represents "now", they're in results[last] to calculate swings.
                results[last][df_cols_last].rename(
                    columns={
                        "total_votes": "total_votes_last",
                        "turnout": "turnout_now_last",
                        "votes": "votes_last",
                        "voteshare": "voteshare_last",
                        "national_polls": "national_polls_now",
                        "geo_polls": "geo_polls_now",
                        "national_voteshare": "national_voteshare_last",
                        "geo_voteshare": "geo_voteshare_last",
                        "winner": "winner_last",
                        "won_here": "won_here_last",
                    }
                ),
                on=["ons_id", "party"],
                how="inner",
                validate="1:1",
            )
            .filter(df_cols_final)
        )
        print(f"Exporting {last}->{now} model dataset to {processed_directory.resolve()}")
        df.to_csv(processed_directory / f"general_election-uk-{now}-model.csv", index=False)

        # Now build dataset to use for predictions
        last = test
        now = pred
        df_cols_last = [
            "ons_id",
            "constituency",
            "county",
            "region",
            "geo",
            "country",
            "electorate",
            "total_votes",
            "turnout",
            "party",
            "votes",
            "voteshare",
            "winner",
            "won_here",
            "national_polls",
            "national_voteshare",
            "national_swing",
            "national_swing_forecast",
            "national_swing_winner",
        ]
        df_cols_final_geo = []
        if "geo_polls" in results[last].columns:
            df_cols_last += ["geo_polls", "geo_voteshare", "geo_swing", "geo_swing_forecast", "geo_swing_winner"]
            df_cols_final_geo += [
                "geo_polls_now",
                "geo_voteshare_last",
                "geo_swing",
                "geo_swing_forecast",
                "geo_swing_winner",
            ]
        df_cols_final = [
            # Constant per constituency
            "ons_id",
            "constituency",
            "county",
            "region",
            "geo",
            "country",
            "electorate",
            "total_votes_last",
            "turnout_last",
            # Constant per party (per constituency)
            "party",
            "votes_last",
            "voteshare_last",
            "winner_last",
            "won_here_last",
            "national_voteshare_last",
            "national_polls_now",
            "national_swing",
            "national_swing_forecast",
            "national_swing_winner",
        ] + df_cols_final_geo
        df = (
            results[last][df_cols_last]
            .rename(
                columns={
                    "total_votes": "total_votes_last",
                    "turnout": "turnout_last",
                    "votes": "votes_last",
                    "voteshare": "voteshare_last",
                    "national_polls": "national_polls_now",
                    "geo_polls": "geo_polls_now",
                    "national_voteshare": "national_voteshare_last",
                    "geo_voteshare": "geo_voteshare_last",
                    "winner": "winner_last",
                    "won_here": "won_here_last",
                }
            )
            .filter(df_cols_final)
        )
        print(f"Exporting {last}->{now} model dataset to {processed_directory.resolve()}")
        df.to_csv(processed_directory / f"general_election-uk-{now}-model.csv", index=False)



import setuptools

with open("README.md", "r") as f:
    long_description = f.read()


setuptools.setup(
    name="maven",
    version="0.0.11",
    description=(
        "Maven's goal is to reduce the time data scientists spend on data cleaning and preparation "
        "by providing easy access to open datasets in both raw and processed formats."
    ),
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords="maven open data etl pipeline",
    author="John Sandall",
    author_email="contact@coefficient.ai",
    url="https://github.com/john-sandall/maven",
    packages=setuptools.find_packages(),
    include_package_data=True,
    install_requires=["pandas==1.0.0", "requests==2.22.0", "xlrd==1.2.0",],
    python_requires="==3.7.*",
    setup_requires=["pytest-runner"],
    test_suite="tests",
    tests_require=["pytest"],
    license="Apache 2.0",
    zip_safe=False,
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Natural Language :: English",
        "Intended Audience :: Developers",
    ],
)



"""
General Election polling data for the United Kingdom.

Usage:
    > import maven
    > maven.get('general-election/UK/polls', data_directory='./data/')

Sources:
    - SixFifty polling data: https://github.com/six50/pipeline/tree/master/data/polls/
        - https://s3-eu-west-1.amazonaws.com/sixfifty/polls.csv
        - https://s3-eu-west-1.amazonaws.com/sixfifty/polls_london.csv
        - https://s3-eu-west-1.amazonaws.com/sixfifty/polls_scotland.csv
        - https://s3-eu-west-1.amazonaws.com/sixfifty/polls_wales.csv
        - https://s3-eu-west-1.amazonaws.com/sixfifty/polls_ni.csv
"""
from pathlib import Path

from maven.datasets.general_election.base import Pipeline


class UKPolls(Pipeline):
    """Handles General Election polling data for the United Kingdom."""

    def __init__(self, directory=Path("data/general-election/UK/polls")):
        super(UKPolls, self).__init__(directory=directory)  # inherit base __init__ but override default directory
        self.sources = [
            # tuples of (url, filename, checksum)
            ("https://s3-eu-west-1.amazonaws.com/sixfifty/", "polls.csv", "98f865803c782e1ffd0cdc4774707ae5"),
            ("https://s3-eu-west-1.amazonaws.com/sixfifty/", "polls_london.csv", "97eb4254039a6bca1a882a9afde2b067"),
            ("https://s3-eu-west-1.amazonaws.com/sixfifty/", "polls_scotland.csv", "096354c852a7c30e22a733eec133b9e3"),
            ("https://s3-eu-west-1.amazonaws.com/sixfifty/", "polls_wales.csv", "2134d55e5288bd5b12be2471f4aacab7"),
            ("https://s3-eu-west-1.amazonaws.com/sixfifty/", "polls_ni.csv", "ea871fad0ce51c03dda09ecec0377dc6"),
        ]
        self.retrieve_all = True
        self.verbose_name = "UKPolls"
